' WeekDayofGDate.bas
' ------------------

' BASIC program to compute the weekday of a Gregorian date specified on the
' command line as D,M,YYYY. (The separator between the date components may be
' a space, comma or slash, but multiple consecutive separators may not be
' used.) If no date is specified, the current system date is used.

' Compiler: PowerBasic Console Compiler 5.5 for Windows

#COMPILE EXE
#DIM ALL

' ---  Program constants  ---

%False = 0
%True  = not %False

' ---  Shared Variables  ---
global YearType&



function PBMain&
'        -------
' Calls: FnYearTypeG&, FnParseDate&, FnWeekdayG&
' Sets Global Variable: YearType&

    local D$, Year&, Month&, Day&, DateValid&, _
          Weekdays$, Weekday$, WeekdayNr&

    D$ = trim$(command$)
    if D$ = "" then
       D$ = Date$  ' if no command-line date, use system date
       Month& = val(left$(D$,2))
       Day&   = val(mid$(D$,4,2))
       Year&  = val(mid$(D$,7,4))
       DateValid& = %True
       YearType& = FnYearTypeG& (Year&)
    else
       DateValid& = FnParseDate& (D$, Day&, Month&, Year&)
    end if

    if isfalse DateValid& then
       print "Invalid Date."
    else
       Weekdays$ = "SunMonTueWedThuFriSat"
       WeekdayNr& = FnWeekdayG& (Year&, Month&, Day&, YearType&)
       ' WeekdayNr& will be 1 to 7 for Sun to Sat
       Weekday$ = mid$(Weekdays$, (WeekdayNr&-1)*3+1, 3)
       print Weekday$
    end if
    print "Press any Key ..." : waitkey$
end function  ' PBMain& -----------------------------------



' FnParseDate& (D$, Day&, Month&, Year&)
' ------------
' From:  PBMain&
' Calls: FnYearTypeG&
' Sets Global Variable: YearType&

' Parses a string (D$) representing a Gregorian date into its component values
' (Day&, Month&, Year&), and returns %True (-1) if it is a valid and correctly
' formatted date, or %False (zero) otherwise. The date must be specified as Day,
' Month, Year, and the year number must be specified in full. The delimiter
' separating the components may be a space, comma or slash (/), but multiple
' consecutive delimeters are not allowed. BC years must be specified using
' astronomical numbering, where years prior to 1 AD are numbered 0, -1, -2, -3,
' etc, in reverse chronological order, so that Y BC = astronomical year 1 - Y.

function FnParseDate& (D$, Day&, Month&, Year&)
'        ------------
    local Delimiter$, Posn&, MaxDays&

    Delimiter$ = ",/ "
    ' Date string must have 2 delimiters at or after positions 2 & 4
    Posn& = instr(2, D$, any Delimiter$)
    if Posn& = 0 then exit function
    Posn& = instr(Posn&+2, D$, any Delimiter$)
    if Posn& = 0 then exit function

    Day&   = val(parse$(command$, any Delimiter$, 1))
    Month& = val(parse$(command$, any Delimiter$, 2))
    Year&  = val(parse$(command$, any Delimiter$, 3))
    YearType& = FnYearTypeG& (Year&)
    ' YearTYpe& will be 0 for common years, -1 for leap years
    select case Month&
      case < 1, > 12   : exit function
      case 2           : MaxDays& = 28 - YearType&
      case 4, 6, 9, 11 : MaxDays& = 30
      case else        : MaxDays& = 31
    end select
    if Day& < 1 or Day& > MaxDays& then exit function
    function = %True
end function  ' FnParseDate&  ---------------------------------------



' FnYearTypeG& (Year&)
' ------------
'   Calls: FnModulo&&

'   Returns %True (-1) if Year& is a Gregorian leap year, otherwise %False (0)

'   BCE years should be specified using astronomical numbering, where years
'   prior to 1 AD are numbered 0, -1, -2, -3, etc, in reverse chronological
'   order, so that Y BC = astronomical year 1 - Y.

function FnYearTypeG& (Year&)
'        ------------
    local  A&, B&, C&

    ' The following lines set A&, B& and C& to either %True or %False
    A& = FnModulo&& (Year&,    4, 0, 0) = 0  ' true if Yr is a multiple of 4
    B& = FnModulo&& (Year&,  100, 0, 0) = 0  ' true if Yr is a multiple of 100
    C& = FnModulo&& (Year&,  400, 0, 0) = 0  ' true if Yr is a multiple of 400

    function = (A& and not B&) or C&
end function  ' FnYearTypeG&  ---------------------------------------



' FnWeekdayG& (Y&, M&, D&, YearType&)
' -----------
'   Calls: FnDayOfYear&, FnModulo&&

'   Gives the weekday (1 to 7, for Sun to Sat) of a Gregorian date,
'   specified as year, month, and day-of-month.
'   Requires the year type (common/leap) as specified below.

'   At call - Y& = the year number
'           - M& = the month number (1-12, for Jan to Dec)
'           - D& = the Day-of-Month
'           - YearType& = 0 for a common year, -1 for a leap year.

'   We compute the total weekday shift (WS) from the known weekday (Sat) of a
'   base date (B) to our date (D). For mathematical convenience, B = 1/1/0,
'   (i.e. 1st January, 1 BCE). For Y > 0, the WS of B to D is the sum of:
'   (a) a yearly component for the WS from B to 1/1/Y, and
'   (b) a Day-of-Year component for the days from 1/1/Y to D.
'   (a) = 1 day per year plus 1 for each occurrence of Feb 29th during the
'   period from from B to D, and (b) = the Day-of-Year - 1.
'   For y < 0, the days (if any) from D back to 1/1/Y are not part of the
'   period from B to D, so we must subtract (b) from (a), but (a) is negative,
'   so the formula is still (a) + (b), as it is for Y > 0. For Y = 0, there is
'   no yearly component, only the (positive) DoY component.

function FnWeekdayG& (Y&, M&, D&, YearType&)
'        -----------
    local  DoY&, Years&, MaxLeaps&, Centuries&, C4&, Leaps&, WeekdayShift&

    DoY& = FnDayOfYear& (YearType&, M&, D&)  ' Day-of-year of our date

    '--- Leap years calculation for the yearly component of the weekday shift:
    if Y& > 0 then      ' For positive years, we
       Years& = Y& - 1  ' don't count Feb 29 of our year (it's counted in DoY).
       Leaps& = 1       ' Includes Feb 29 of year 0 (not part of calc'n below).
    else                ' But for negative years, the (backward) weekday shift
       Years& = Y&      ' from B to 1/1/y does include Feb 29 of year y.
    end if
    ' Calculate the nr of leap years from year 1 (or -1) to year y:
    MaxLeaps&     = fix(Years&/4)      ' every 4th year is leap
    Centuries&    = fix(Years&/100)    ' except century years, unless ...
    C4&           = fix(Centuries&/4)  ' the century yr is a multiple of 400
    Leaps&        = Leaps& + MaxLeaps& - Centuries& + C4&

    ' Calculate the total weekday shift from the known wkday of a base date:
    ' Our base date (Jan 1, 0) is Sat = day 7 = day 0. (No need to add zero.)
    WeekdayShift& = Y& + Leaps& + DoY& - 1
    ' Adjusted Modulo result is 7 instead of 0 for Saturdays.
    function = FnModulo&& (WeekDayShift&, 7, -1, 0)
end function  ' FnWeekDayG&  ----------------------------------------



' FnDayOfYear& (YearType&, M&, D&)
'  ------------
'   From : FnWeekdayG&
'   Calls: FnEndOfMonth&

'   Gives the Day-of-Year (1 to 366) of a date, specified as month and
'   day-of-month, for the given year-type (common or leap), i.e. the
'   number of days elapsed in the year up to and including that date.
'   This is sometimes ambiguously called the Julian date, which invites
'   confusion with dates of the Julian calendar or with the Julian Day Number,
'   a day count used in astronomy and calendar computations that is similar to
'   this but which spans years. The unambiguous term Day-of-Year is preferrable.

'   At call - YearType& = 0 for a common year, -1 for a leap year.
'           - M& = the month number (1-12, for Jan to Dec)
'           - D& = the Day-of-Month

function FnDayOfYear& (YearType&, M&, D&)
'        ------------
    local  EndOfPreviousMonth&

    EndOfPreviousMonth& = FnEndOfMonth& (M&-1, YearType&)
    function = EndOfPreviousMonth& + D&
end function  ' FnDayOfYear&  ---------------------------------------



' FnEndOfMonth& (M&, YearType&)
' -------------
'   From : FnDayOfYear&

'   Gives the Day-of-Year (31 to 366) of the last day of month M& for a given
'   year-type (common or leap), i.e. the number of days elapsed in the year, at
'   the end of month M&. If M& = 0, the result returned is zero. This is useful
'   for a procedure that calls this function to obtain the DoY at the end of
'   the previous month (i.e. CurrentMonth minus 1). In such a case, this will
'   work even when the current month is month 1 (January).

'   At call - M& = the month number (1-12, for Jan to Dec)
'           - YearType& = 0 for a common year, -1 for a leap year.

'   We multiply the month nr by Motti's Constant (30.57), an approximate,
'   average month-length. The rounded product, decremented, for months > 1,
'   by an adjustment factor, is the Day-of-Year at the end of that month.
'   (The fraction in the product is never 0.5, so "banker's rounding" is never
'   applicable.) The adjustment factor is 2 for a common year, 1 for a leap yr.

' Month:      1    2    3     4     5     6     7     8     9    10    11    12
' Product:  30.6 61.1 91.7 122.3 152.9 183.4 213.9 244.6 275.1 305.7 336.3 366.8
' Rounded:  31   61   92   122   153   183   214   245   275   306   336   367
' Adjusted: 31   59   90   120   151   181   212   243   273   304   334   365

function FnEndOfMonth& (M&, YearType&)
'        -------------
    local  MottisConstant!, Adjust&

    MottisConstant! = 30.57   ' = approximate average month length
    if M& < 2 then Adjust& = 0 else Adjust& = YearType& + 2
    ' CINT is not strictly needed. Without it, PB would round implicitly:
    function = cint(M& * MottisConstant!) - Adjust&
end function  ' FnEndOfMonth&  --------------------------------------



' FnModulo&& (Dividend&&, Divisor&, Adjust&, IntQuotient&)
' ----------
'   Gives the integer quotient (Q) and the remainder (R) of the integer
'   division, Dividend / Divisor where both dividend and divisor are integers.
'   R is the function's return value and Q is passed back in the fourth
'   parameter, IntQuotient. If you only want the remainder, pass zero (or any
'   numeric literal) as the fourth argument. (Its value at call is ignored.)

'   If parameter 3, Adjust&, is non-zero, the function returns ABS(Divisor)
'   instead of 0 when there is no remainder. (See explanation of this below.)

'   NOTE: The remainder (R) returned by this function is the same as the result
'   of the inbuilt modulo function on HP scientific calculators of the early
'   1980s. It sometimes differs from the result (M) of BASIC's intrinsic MOD
'   function. The difference lies in the method of converting the real quotient
'   (q) to an integer (Q).

'   BASIC's MOD function converts q to Q by truncating q. The decimal portion
'   is discarded, leaving only the integer portion. This endows the MOD
'   function with the properties that:
'   (1) for Dividends < 0, M will be <= 0, for all Divisors, >0 or <0.
'       Thus, for non-zero M, SGN(M) = SGN(Dividend)
'   (2) ABS(M) will be the same as ABS(Dividend) MOD ABS(Divisor).

'   Q is defined here as the largest integer less than or equal to q (i.e. q
'   is rounded in the direction of minus infinity), e.g. for q = 3.9, Q = 3,
'   for q = -2.2, Q = -3. In other languages, this method of conversion to
'   integer is called the Floor function.

'   This method endows this function with two important properties:
'   (1) For non-zero R, SGN(R) = SGN(Divisor), thus,
'       for Divisors > 0, R will be => 0 for all dividends, even negative ones.
'   (2) when SGN(Dividend) =  SGN(Divisor), R = M, but
'       when SGN(Dividend) <> SGN(Divisor), R = M + Divisor.

'   These two properties ensure sensible results for all modular arithmetic
'   (especially subtraction) performed on members of a set of finitely-numbered
'   items that are numbered 'modularly'. That is, after the highest number has
'   been reached, the item numbering wraps around and starts again from the
'   first number for subsequent members of the set. Taking weekday numbering as
'   an example of this, this function correctly computes backward weekday shifts
'   whereas MOD will not. Assuming Sunday = weekday 1, you can find the weekday
'   8 days before Friday (weekday 6) from the remainder of the division (6-8)/7.
'   With this function, the result is 5 (Thursday); with MOD, you get -2 which
'   then requires you to go through the additional step of summing the MOD
'   result and the divisor to obtain the answer in a meaningful form.

'   The optional feature whereby ABS(Divisor) is returned instead of zero when
'   there is no remainder is useful for modular arithmetic on items numbered
'   modularly as described above, but where the numbering commences with 1
'   rather than zero, e.g. where weekdays are numbered 1 to 7 for Sunday to
'   Saturday. A result indicating Saturday will then be 7 instead of zero.

function FnModulo&& (byval Dividend&&, byval Divisor&, byval Adjust&, byref IntQuotient&)
'        ----------
    local Remainder&&

    IntQuotient& = int(Dividend&&/Divisor&)
    Remainder&&  = Dividend&& - (IntQuotient& * Divisor&)
    function = Remainder&&

    if istrue Adjust& and Remainder&& = 0 then
       function = abs(Divisor&)
    end if
end function  ' FnModulo&&  -----------------------------------------
